//
// Simple demo for GL_EXT_fog_coord OpenGL extension
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru
//

#include	<windows.h>
#include	<stdio.h>
#include	<gl/gl.h>
#include	<gl/glut.h>
#include	<gl/glaux.h>
#include	"../glext.h"

#include	"Vector3D.h"
#include	"Vector2D.h"

extern "C"
{
	#include	"tga.h"
}

bool	loadTexture    ( int target, bool mipmap, const char * fileName );
bool	loadTgaTexture ( int target, bool mipmap, const char * fileName );
bool	loadBmpTexture ( int target, bool mipmap, const char * fileName );

PFNGLACTIVETEXTUREARBPROC			glActiveTextureARB       = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC 	glClientActiveTextureARB = NULL;
PFNGLMULTITEXCOORD1FARBPROC			glMultiTexCoord1f        = NULL;
PFNGLMULTITEXCOORD1FVARBPROC		glMultiTexCoord1fv       = NULL;
PFNGLMULTITEXCOORD2FARBPROC			glMultiTexCoord2f        = NULL;
PFNGLMULTITEXCOORD2FVARBPROC		glMultiTexCoord2fv       = NULL;
PFNGLFOGCOORDFPROC					glFogCoordf              = NULL;


#define	N1	40
#define	N2	20

struct	Facet
{
	int	 index [3];					// indices into vertex, normal and tex arrays
};

Vector3D	vertex [N1 * N2];
Vector3D	normal [N1 * N2];
Vector2D	tex    [N1 * N2];
Facet		facet  [N1 * N2 * 2];

float	 angle  = 0;
float	 angle2 = 0;
unsigned texture1;					// main texture

Vector3D	eye      ( 0, 0, 0  );
Vector3D	lookAt   ( 0, -5, 5 );
Vector3D	up       ( 0, 1, 0  );
Vector3D	fogColor ( 0, 0.7f, 0.8f );

void	initTorus ( float r1, float r2 )
{
					// 1. Create vertices
	for ( int i = 0, k = 0; i < N1; i++ )
	{
		float	phi = i * 2 * M_PI / N1;

		for ( int j = 0; j < N2; j++, k++ )
		{
			float	psi = j * 2 * M_PI / N2;

											// compute vertex coordinates
			float	temp = r1 + r2 * cos ( psi );
			float	dt   =     -r2 * sin ( psi );

			vertex [k].x = temp * cos ( phi );
			vertex [k].y = temp * sin ( phi );
			vertex [k].z = r2 * sin ( psi );

											// compute texture coordinates
			tex [k].x = (float) i / (float) N1;
			tex [k].y = (float) j / (float) N2;

											// compute normals
			Vector3D	du ( -temp * sin ( phi ), temp * cos ( phi ), 0 );
			Vector3D	dv ( dt * cos ( phi ), dt * sin ( phi ), r2 * cos ( psi ) );

			normal [k] = (du ^ dv).normalize ();
		}
	}

					// 2. Create facets
	for ( i = k = 0; i < N1; i++ )
		for ( int j = 0; j < N2; j++, k += 2 )
		{
			facet [k].index [0] = i*N2 + j;
			facet [k].index [1] = ((i+1) % N1)*N2 + j;
			facet [k].index [2] = ((i+1) % N1)*N2 + (j + 1) % N2;

			facet [k+1].index [0] = i*N2 + j;
			facet [k+1].index [1] = ((i+1) % N1)*N2 + (j + 1) % N2;
			facet [k+1].index [2] = i*N2 + (j + 1) % N2;
		}
}

void	drawTorus ()
{
	float	mvMatrix [16];

						// get modelview matrix to transform back
	glGetFloatv ( GL_MODELVIEW_MATRIX, mvMatrix );


	glBegin ( GL_TRIANGLES );

	for ( int i = 0; i < N1 * N2 * 2; i++ )
		for ( int j = 0; j < 3; j++ )
		{
			int	 k   = facet [i].index [j];
			Vector3D v   = vertex [k];
			float    fog = -v.x * mvMatrix [1] - v.y * mvMatrix [5] - v.z * mvMatrix [9] - mvMatrix [13];

			glFogCoordf   ( fog        );
			glNormal3fv   ( normal [k] );
			glTexCoord2fv ( tex    [k] );
			glVertex3fv   ( vertex [k] );
		}

	glEnd ();
}

unsigned	createTexture2D ( bool mipmap, const char * fileName )
{
	unsigned	textureId;

	glGenTextures   ( 1, &textureId );
	glBindTexture   ( GL_TEXTURE_2D, textureId );
	glPixelStorei   ( GL_UNPACK_ALIGNMENT, 1 );							// set 1-byte alignment

	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );	// set texture to repeat mode
	glTexParameteri ( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

	loadTexture ( GL_TEXTURE_2D, mipmap, fileName );

	glTexParameteri   ( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );	
	glTexParameteri   ( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );

	return textureId;
}

bool	loadTexture ( int target, bool mipmap, const char * fileName )
{
	const char * ext = strrchr ( fileName, '.' );

	if ( ext == NULL )
		return false;

	if ( !stricmp ( ext + 1, "bmp" ) )
		return loadBmpTexture ( target, mipmap, fileName );
	else
	if ( !stricmp ( ext + 1, "tga" ) )
		return loadTgaTexture ( target, mipmap, fileName );

	printf ( "oops\n" );

	return false;
}

bool	loadTgaTexture ( int target, bool mipmap, const char * fileName )
{
	FILE            * file;
	gliGenericImage * image;

	file = fopen ( fileName, "rb" );

	if ( file == NULL )
	{
		printf ( "Error loading %s.\n", fileName );

		return false;
	}

	image = gliReadTGA ( file, (char *)fileName );

	fclose ( file );

	if ( mipmap )
		gluBuild2DMipmaps ( target, image -> components,
			                image -> width, image -> height,
							image -> format, GL_UNSIGNED_BYTE, image -> pixels );
	else
	    glTexImage2D (      target, 0, image -> components,
				            image -> width, image -> height, 0,
							image -> format, GL_UNSIGNED_BYTE, image -> pixels );
	return true;
}

bool	loadBmpTexture ( int target, bool mipmap, const char * fileName )
{
								// load texture
	AUX_RGBImageRec * texture = auxDIBImageLoad ( fileName );

	if ( texture == NULL )
	{
		printf ( "Error loading %s.\n", fileName );

		return false;
	}

	if ( mipmap )
		gluBuild2DMipmaps ( target, GL_RGB, 
			                texture -> sizeX, texture -> sizeY, 
				            GL_RGB, GL_UNSIGNED_BYTE, 
					        texture -> data );
	else
	    glTexImage2D ( target, 0, GL_RGB,
					   texture -> sizeX, texture -> sizeY, 0,
					   GL_RGB, GL_UNSIGNED_BYTE, texture -> data );


	return true;
}

bool	isExtensionSupported ( const char * ext )
{
	const char * extensions = (const char *)glGetString ( GL_EXTENSIONS );
	const char * start      = extensions;
	const char * ptr;

	while ( ( ptr = strstr ( start, ext ) ) != NULL )
	{
											// we've found, ensure name is exactly ext
		const char * end = ptr + strlen ( ext );

		if ( isspace ( *end ) || *end == '\0' )
			return true;

		start = end;
	}

	return false;
}

bool	initExtensions ()
{
	glActiveTextureARB       = (PFNGLACTIVETEXTUREARBPROC)       wglGetProcAddress ( "glActiveTextureARB"       );
	glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC) wglGetProcAddress ( "glClientActiveTextureARB" );
	glMultiTexCoord1f        = (PFNGLMULTITEXCOORD1FARBPROC)     wglGetProcAddress ( "glMultiTexCoord1fARB"     );
	glMultiTexCoord1fv       = (PFNGLMULTITEXCOORD1FVARBPROC)    wglGetProcAddress ( "glMultiTexCoord1fvARB"    );
	glMultiTexCoord2f        = (PFNGLMULTITEXCOORD2FARBPROC)     wglGetProcAddress ( "glMultiTexCoord2fARB"     );
	glMultiTexCoord2fv       = (PFNGLMULTITEXCOORD2FVARBPROC)    wglGetProcAddress ( "glMultiTexCoord2fvARB"    );
	glFogCoordf              = (PFNGLFOGCOORDFPROC)              wglGetProcAddress ( "glFogCoordfEXT"           );


	return glActiveTextureARB       != NULL &&
		   glClientActiveTextureARB != NULL &&
		   glMultiTexCoord1f        != NULL &&
		   glMultiTexCoord1fv       != NULL &&
		   glMultiTexCoord2f        != NULL &&
		   glMultiTexCoord2fv       != NULL &&
		   glFogCoordf              != NULL;   
}

void init ()
{
	glClearColor ( 0, 0, 0, 1 );
	glEnable     ( GL_DEPTH_TEST );
	glDepthFunc  ( GL_LEQUAL );

											// main texture
	glEnable ( GL_TEXTURE_2D );

											// setup fog
	glEnable ( GL_FOG                 );
	glFogi   ( GL_FOG_MODE, GL_LINEAR );
	glFogfv  ( GL_FOG_COLOR, fogColor );
	glFogf   ( GL_FOG_START, -1       );
	glFogf   ( GL_FOG_END,   3        );
	glFogi   ( GL_FOG_HINT, GL_NICEST );
	glFogi   ( GL_FOG_COORDINATE_SOURCE_EXT, GL_FOG_COORDINATE_EXT );
}

void display ()
{
	glClear         ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glPushMatrix    ();
	glTranslatef    ( 0, -7, 7 );
	glRotatef       ( angle,  1, 1, 0 );
	glRotatef       ( angle2, 0, 1, 1 );

	drawTorus ();

	glPopMatrix     ();
	glutSwapBuffers ();
}

void reshape ( int w, int h ) 
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,			// eye
					lookAt.x, lookAt.y, lookAt.z,	// center
					up.x, up.y, up.z );
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )	// quit requested
		exit ( 0 );
}

void	animate ()
{
	angle  = 0.04f * glutGet ( GLUT_ELAPSED_TIME );
	angle2 = 0.02f * glutGet ( GLUT_ELAPSED_TIME );

	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								// initialise glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );   
	glutInitWindowSize  ( 500, 500 );


								// create window
	int	glWin = glutCreateWindow ( "OpenGL fog_coord example" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key );
	glutIdleFunc     ( animate );

	const char * vendor    = (const char *)glGetString ( GL_VENDOR     );
	const char * renderer  = (const char *)glGetString ( GL_RENDERER   );
	const char * version   = (const char *)glGetString ( GL_VERSION    );
	const char * extension = (const char *)glGetString ( GL_EXTENSIONS );

	printf ( "Vendor:   %s\nRenderer: %s\nVersion:  %s\n", vendor, renderer, version );

	if ( !isExtensionSupported ( "GL_EXT_fog_coord" ) )
	{
		printf ( "GL_EXT_fog coord NOT supported.\n" );

		return 1;
	}

	if ( !initExtensions () )
	{
		printf ( "initExtensions FAILED.\n" );

		return 2;
	}

	texture1 = createTexture2D ( true, "..\\Textures\\block.bmp" );

	init      ();
	initTorus ( 3, 1 );

	glutMainLoop ();

	return 0;
}
